/********************************************************************
Beschreibung:
Zwei Pollin-Evaluationboards mit jeweils einem RFM12 um miteineander 
zu kommunizieren. Auf den Boards wurden ATmega8 mit 8 MHz internen 
Takt eingesetzt.

Dieser Code ist auf beiden Board drauf.

Pinbelegung
-----------------------
ATMEGA8           RFM12
-----------------------
SCK               SCK
MISO              SDO
MOSI              SDI
SS                nSEL
PD2 (INT0)        nIRQ
PD5               LED1
PD6               LED2
VDD               VDD
PD7               FSK
PD3 (INT1)        FFIT
PB0               VDI
**********************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdlib.h>
#include <compat/deprecated.h>
#include "i2cmaster.h"
#include <string.h>

typedef union conver2_ 
{
	unsigned int w;
	unsigned char b[2];
} CONVERTW;

#define F_CPU 16000000UL
#define BAUD 19200UL

#define LED1_ON		PORTD |= (1 << PD5)
#define LED2_ON		PORTD |= (1 << PD6)
#define LED1_OFF	PORTD &= ~(1 << PD5)
#define LED2_OFF	PORTD &= ~(1 << PD6)
#define LED1_TOG	PORTD ^= (1 << PD5)
#define LED2_TOG	PORTD ^= (1 << PD6)

#define Taster3		PIND  &  (1<<PIND4)

// Funktionsprototypen
void rfm12_write_data(uint16_t);			
unsigned short  rfm12_write(uint16_t);
uint8_t rfm12_read();	
void port_init();		
void spi_master_init();						
void rfm12_init();						
void rfm12_rx_on();
void rfm12_tx_on();
void uart_init();
uint8_t uart_get_byte();
void uart_send_byte(uint8_t);
void uart_send_int(char*);
void rfm_send_text(char*);
unsigned char parity(unsigned char);

uint8_t rfm_packet_len, rfm_empfangsbyte;	// Globale Variablen


volatile uint8_t rfm12_fifo_voll;			// ISR Variable
volatile int counter=0; 					// ISR Variable
volatile unsigned short z=0;				// ISR Variable


unsigned short rfm_status;
int count = 0;
uint8_t b=0;
float korrekturfaktor = 1.00009; 									// Quarzfrequenz hier anpassen fr Abgleich (z.b 0.99974)
int ztmp = 0;
unsigned long freq = 0;


char Text_Array_RH[17];	
char Text_Array_FR[17];
char Text_Array[60];


float sens = 0.0;			
float senslsb = 0.0;
float offset = 0.0;			
float offlsb = 0.0;			
float rh = 0.0;											// Relative Luftfeuchtigkeit

uint8_t SensorEnable=1;

int main(void)
{	
	port_init();											// Ein/Ausgnge initialisieren
	spi_master_init();										// SPI initialisieren
	uart_init();											// USART initialisieren
	rfm12_init();											// RFM12 initialisieren													// Interrupts aktivieren



	if (SensorEnable==1)
	{
		i2c_init();												// IC initialisieren
		i2c_start( 0xA2 );										// IC bertragung starten mit 0xA2
		i2c_write( 0x0A );										// Adresse die als erstes Ausgelesen werden soll
		i2c_start( 0xA3 );										// Daten anfordern
		sens = i2c_readAck();									// 0x0A MSB von sens
		senslsb = i2c_readAck();								// 0x0B LSB von sens
		offset = i2c_readAck();									// 0x0C MSB von offset
		offlsb = i2c_readNak();									// 0x0D LSB von offset, es werden keine weiter Daten mehr angefordert
		i2c_stop();												// IC bertragung beenden

		offset = offset * ( 1<<8 );								// Stellenwert von offset-MSB bercksichten
		offset = offset + offlsb;								// MSB und LSB von Offset zusammenfgen
		sens = sens * ( 1<<8 );									// Stellenwert von sens-MSB bercksichten
		sens = sens + senslsb;									// MSB und LSB von sens zusammenfgen
	}

	sei();
	
	rfm12_rx_on();											// Empfnger einschalten
	
	while(1)												// Arbeitsschleife
	{
		if(rfm12_fifo_voll==1)
		{
			rfm12_fifo_voll=0;								// Interrupt-Flag zurcksetzen
		  	
			rfm_packet_len=rfm12_read();					// Erstes Nutzerbyte ist die Packetlnge

			for(uint8_t i=1;i<=rfm_packet_len;i++)			// Restlichen Nutzerbytes ...
			{
				rfm_empfangsbyte=rfm12_read();				// ... auslesen	
				uart_send_byte(rfm_empfangsbyte);			// Nutzerbyte ber UART an den PC schicken
				LED2_TOG;
			}
			rfm12_write(0x0000);							// Status Register lesen (Interrupts werden gelscht)
			rfm12_write(0xCA81);							// FIFO deinitialisieren
			rfm12_write(0xCA83);							// FIFO deaktivieren
		}


		if (counter==125)
		{ 
			count = TCNT0;
			counter=0;
			ztmp = z;
			z= 0;
			TCNT0 = 0;

			b++;
		}

		if(b==1)
		{
			if (SensorEnable==1)
			{

				rfm_status = rfm12_write(0x0000);

			  	freq = 0.5+((256UL*ztmp + count)*korrekturfaktor); 			
			
				rh = ( offset - freq );
				rh = rh * sens;
				rh = rh / 4096;

				dtostrf( rh, 6, 2, Text_Array_RH);
				

				
				strcpy( Text_Array, Text_Array_RH );
 				strcat( Text_Array, "%  " );

				itoa( freq, Text_Array_FR,10);
 				strcat( Text_Array, Text_Array_FR );

 				strcat( Text_Array, " Hz  " );

				itoa( offset, Text_Array_FR,10);
 				strcat( Text_Array, Text_Array_FR );


 				strcat( Text_Array, "  " );

				itoa( sens, Text_Array_FR,10);
 				strcat( Text_Array, Text_Array_FR );

 				strcat( Text_Array, "  " );

				itoa( rfm_status, Text_Array_FR,2);
 				strcat( Text_Array, Text_Array_FR );
				
				uart_send_int(Text_Array_FR);
				uart_send_int("\n");

 				strcat( Text_Array, "\n\0" );


				rfm_send_text(Text_Array);
				
			}
			b=0;
		}
	}
}


void port_init()
{
	DDRD |= (1 << PD5) | (1 << PD6);						// PD5 (Led1), PD6 (Led2) als Ausgang		
	DDRD  &= ~(1<<PIND4);									// PD4 (Taster3) als Eingang
	
	DDRD  |=  (1<<PD7);										// PD7 (nFFS) als Ausgang
	PORTD |=  (1<<PD7);										// nFFS auf high ziehen
	
	DDRD  &= ~(1<<PIND2);					 				// PD2(INT0) fr den nIRQ-Pin	
	DDRD  &= ~(1<<PIND3);				 					// PD3(INT1) fr den FIFO-Interrupt

	MCUCR |= (1<<ISC10) | (1<<ISC11);						// INT1 steigende Flanke
	GICR  |= (1<<INT1);										// INT1 aktivieren


	TCCR2 |= (1<<WGM21); 									// CTC Mode Aktivierung Timer2
	TCCR2 |= (1<<COM20) | (1<<COM21); 						// set OCR Flag bei "Compare Match"
	TCCR0 |= (1<<CS02) | (1<<CS01) | (1<<CS00);			 	// Ext Flanke Interrupt (T0)
	
	
	// 8 MHz
	// TCCR2 |= (1<<CS22) | (1<<CS21);				 		// Prescaler Timer2 auf 256 setzen
	// OCR2  |= 249; 										// Output Compare register Timer2 auf 249 (fr Loop 0-249 => 250)

//  16 MHz
	TCCR2 |= (1<<CS22) | (1<<CS21) | (1<<CS20);				// Prescaler Timer2 auf 1024 setzen
	OCR2  |= 124; 											// Output Compare register Timer2 auf 124 (fr Loop 0-124 => 125)

	TIMSK |= (1<<OCIE2); 									// Enable comp match flag interrupt Timer 2 
	TIMSK |= (1<<TOIE0); 									// Enable overflow flag interrupt Timer 0
}

void spi_master_init (void) 
{
	uint8_t status;
	DDRB = (1<<PB2) | (1<<PB3) | (1<<PB5);					// setze SCK,MOSI,PB2 (SS) als Ausgang
	DDRB &= ~(1<<PB4);										// setze MISO als Eingang
	PORTB = (1<<PB5) | (1<<PB2);							// SCK und PB2 high (ist mit SS am Slave verbunden)
	SPCR = (1<<SPE) | (1<<MSTR) | (1<<SPI2X)  | (1<<SPR0);	// Aktivierung des SPI, Master, Taktrate fck/8
	status = SPSR;											// Status lschen
}


// ***********************************************************************************************
// ** 									Interrupt-Routinen 										**
// ***********************************************************************************************

ISR(INT1_vect)												// FIFO-Interrupt (PIN: FFIT)
{
	rfm12_fifo_voll=1;										// Interrupt-Flag setzen
}

ISR(TIMER2_COMP_vect) // Lst aus alle 8ms aus (16Mhz/1024/125)
{
	counter++;
}

ISR(TIMER0_OVF_vect) // Eingangsignal Overflow Zhler Timer 0
{
	z++;
	TIFR = (1<<TOV0); //ov flag timer 0 zurcksetzen
}

// ***********************************************************************************************
// ** 									RFM12-Funktionen										**
// ***********************************************************************************************

void rfm12_init(void)
{
	for(uint8_t t=0;t<20;t++)_delay_ms(10);
  	rfm12_write(0x80D8);									// TX-FIFO, RX-FIFO, Band = 433 MHz, Quarz = 12,5 pF 	
	rfm12_write(0x82D1);									// Empfnger, Basis-Band, Synthesizer, Kein Takt-Ausgang
  	rfm12_write(0xA140);									// Sendefrequenz = 430.8 MHz
	rfm12_write(0xC623);									// 9600 Bit/s
	rfm12_write(0x9480);									// VDI, Schnell, 200 KHz, 0 dB, -103 dB
  	rfm12_write(0xC2AC);									// Auto-Lock, Digitalfilter, mittlere Schwelle
	rfm12_write(0xCA81);									// FIFO = 8 Bit, Synchronmuster, FIFO leeren/sperren = Empfnger, Reset-Schaltung aus
  	rfm12_write(0xC4F7);									// Auto-Mode = VDI unabhngig, Range-Limit = -4/+3, Fine-Mode, Offset aktiv, AFC aktiv
  	rfm12_write(0x9800);									// Modulation-Polarity = normal, Frequenzhub = 15 KHz, Ausgangsleistung = Maximal
  	//rfm12_write(0x0000);									// Status Register lesen (Interrupts werden gelscht)
	rfm12_write(0xCA81);									// FIFO initialisieren
	rfm12_write(0xCA83);									// FIFO aktivieren
}

void rfm12_write_data(uint16_t wert)
{
	while(PIND & (1<<PIND2));								// Warten bis Modul bereit
	rfm12_write(wert);										// Sendeword bertragen
}

unsigned short  rfm12_write(uint16_t wert)
{
/*
	uint16_t newHigh=0,newLow=0,newWord=0;				
	PORTB &= ~(1<<PB2);										// SS am Slave Low --> Beginn der bertragung
	SPDR = wert >> 8;										// High-Byte bertragen
	while (!(SPSR & (1<<SPIF)));							// Warten bis bertragung abgeschlossen
	newHigh = SPDR;											// Empfangenes High-Byte lesen
	SPDR = wert;											// Low-Byte bertragen
	while (!(SPSR & (1<<SPIF)));							// Warten bis bertragung abgeschlossen
	newLow = SPDR;											// Empfangenes Low-Byte lesen
	PORTB |= (1<<PB2);										// SS am Slave High --> Ende der bertragung
	newWord = (newHigh<<=8) | newLow;						// High-Byte und Low-Byte zusammenfhren	
	return newWord;											// Rckgabe des empfangenen Words
*/
	CONVERTW val;
	val.w=wert;
	cbi(PORTB, PB2);
	SPDR = val.b[1];
	while(!(SPSR & (1<<SPIF)));
	val.b[1]=SPDR;
	SPDR = val.b[0];
	while(!(SPSR & (1<<SPIF)));
	val.b[0]=SPDR;
	sbi(PORTB, PB2);
	return val.w;

}

uint8_t rfm12_read(void)
{
	uint16_t FIFO_data;							
	//rfm12_write(0x0000);									// Status Register lesen (Interrupts werden gelscht)
	while(PIND & (1<<PIND2));								// Warten bis Modul bereit
	FIFO_data=rfm12_write(0xB000);							// FIFO-Daten einlesen (16Bit)
	return(FIFO_data&0x00FF);								// LOW-Byte der FIFO-Daten zurckgeben (Datenbyte)
}

void rfm12_rx_on()
{
	rfm12_write(0x82C8);									// Empfnger einschalten
	_delay_ms(1);											// Warten bis Empfnger an ist
	rfm12_write(0xCA81);									// FIFO initialisieren
	rfm12_write(0xCA83);									// FIFO aktivieren
}

void rfm12_tx_on()
{
	rfm12_write(0x8208);									// Empfnger ausschalten
	_delay_ms(1);											// Warten bis Empfnger aus ist
	//rfm12_write(0x0000);									// Status Register lesen (Interrupts werden gelscht)
	rfm12_write(0x8271);									// Sender aktivieren
}

void rfm_send_text(char *s)
{
	uint8_t i = 0;											// Zhler fr Packetlnge
	char *a = s;											// Adresse von Pointer in a sichern
	while(*s)												// Speicher bis Packetende durchlaufen
	{
		i++;												// Packetlnge mitzhlen (ermitteln)
		s++;												// Nchstes Nutzerbyte whlen
	}
	
	rfm12_tx_on();											// Sender einschalten
	rfm12_write_data(0xB8AA);								// Prambel senden			
	rfm12_write_data(0xB8AA);								// Prambel senden						
	rfm12_write_data(0xB8AA);								// Prambel senden						
	rfm12_write_data(0xB82D);								// Syncronmuster senden			
	rfm12_write_data(0xB8D4);								// Syncronmuster senden				
	
	rfm12_write_data(0xB800 | i);							// Nutzerbyte senden							

	s=a;													// Pointeradresse wieder aus a laden
	
	while(*s) 												// Speicher bis Packetende durchlaufen
	{
		rfm12_write_data(0xB800 | *s);						// Nutzerbyte senden
		s++;
		LED2_TOG;
	}
	rfm12_write_data(0xB8AA);								// Dummybyte
	rfm12_write_data(0x8208);								// Sender ausschalten
	rfm12_rx_on();											// Empfnger einschalten
}

// ***********************************************************************************************
// ** 									USART-Funktionen										**
// ***********************************************************************************************

ISR(USART_RXC_vect)											// USART Empfangs-Interrupt
{
	uint8_t daten;
	daten = uart_get_byte();								// Zeichen aus Puffer holen
}

void uart_init(void)
{
	uint8_t x;
	UBRRL = (F_CPU / (16ul * BAUD))-1;						// Baudrate mit Takt und Baud
	UCSRB |= (1 << TXEN) | (1 << RXEN); 					// Sender und Empnger aktivieren
	UCSRC |= (1 << URSEL) | (1 << UCSZ1) | (1 << UCSZ0);	// async 8bit
	x = UDR;												// Empfnger leeren
	sbi(UCSRB,RXCIE );										// USART-Empfangs-Interrupt aktivieren
}

uint8_t uart_get_byte(void)
{
	while(!(UCSRA & (1 << RXC)));							// warten bis Empfang abgeschlossen
	return UDR;												// Byte abholen
}

void uart_send_byte(uint8_t x)
{
	while(!(UCSRA & (1 << UDRE)));							// warten bis Schnittstelle bereit
	UDR = x;												// Zu sendendes Byte bergeben
}

void uart_send_int(char *s)
{
    while (*s)												// so lange *s ungleich '\0' (String-Endezeichen)
    {   
        uart_send_byte(*s);									// Byte senden
        s++;												// nchstes Byte whlen
    }
	//uart_send_byte('\r');									// Carriage Return senden
}
